//===================================================== file = bqueue1.c =====
//=  A CSIM18 simulation of queue with block transfers                       =
//=   - Single input port queue, single crossbar queue                       =
//============================================================================
//=  Notes: This model is to serve as a guide for implementing a RR/RR CICQ  =
//=         switch with block transfers                                      =
//=--------------------------------------------------------------------------=
//=  History:   KJC (01/17/02) - Genesis                                     =
//============================================================================
#pragma warn -pro           // Disable function prototype warning (bcc)
#pragma warn -stu           // Disable undefined structure warning (bcc)

//----- Includes -------------------------------------------------------------
#include <assert.h>         // Needed for assert() macro
#include <stdio.h>          // Needed for printf()
#include <math.h>           // Needed for sqrt()
#include "csim.h"           // Needed for CSIM18 stuff

//----- Defines --------------------------------------------------------------
#define      FALSE       0  // Boolean false
#define       TRUE       1  // Boolean true
#define MEDIA_RATE  10.0e6  // Media bit rate (bits per millisec)
#define BLOCK_SIZE    1500  // Block transfer size in bytes
#define THRESHOLD     1500  // Threshold for transfer in CP buffer
#define STATUS_LEN      64  // Time for status reporting (byte times)

//----- Globals --------------------------------------------------------------
EVENT    Q_tail;            // Queueing event for inport queue tail
EVENT    Q_head;            // Queueing event for inport queue head
EVENT    Below_thresh;      // Event for below transfer threshold
FACILITY Server;            // Server facility for outgoing link
TABLE    Delay_table;       // Delay table for packets
int      CP_buf_len;        // CP buffer length
int      HOL_frag_len;      // HOL fragment length
double   Lambda;            // Mean arrival rate (pkt/sec)
double   Sim_time;          // Simulation time (in millisec)

//----- Prototypes -----------------------------------------------------------
void generate(void);                        // Generate IEEE mix packets
void inport(int pkt_len, double org_time);  // Input port with inport queue
void outport(int pkt_len, double org_time); // Output port with CP buffer
void poll(void);                            // Block poller
void status_reporter(void);                 // Status reporter

//============================================================================
//==  Main program                                                          ==
//============================================================================
void sim(void)
{
  // Create the simulation
  create("sim");

  // Increase maximum number of allowed processes
  max_processes(1000000L);

  // CSIM initializations
  Q_tail = event("Inport queue tail");
  Q_head = event("Inport queue head");
  Below_thresh = event("CP is below transfer threshold event");
  Server = facility("Server for output link");
  Delay_table = table("Delay table for packet delay");

  // Other initializations
  set(Q_tail);
  CP_buf_len = 0;
  Sim_time = 50.0;
  Lambda = 0.95 * 2037.489;

  // Output begin-of-simulation banner
  printf("*** BEGIN SIMULATION *** \n");

  // Initiate generate, poll, and status_reporter functions and then simulate
  generate();
  poll();
  status_reporter();
  hold(Sim_time);

  // Output results
  printf("============================================= bqueue1.c ===== \n");
  printf("== >>> CSIM18 simulation of queue with block transfers <<< == \n");
  printf("============================================================= \n");
  printf("= Lambda                = %9.3f pkt/millisec  \n", Lambda);
  printf("= BLOCK_SIZE            = %d bytes            \n", BLOCK_SIZE);
  printf("= THRESHOLD (in CP)     = %d bytes            \n", THRESHOLD);
  printf("= STATUS_LEN            = %d byte times       \n", STATUS_LEN);
  printf("============================================================= \n");
  printf("= Total CPU time        = %9.3f sec           \n", cputime());
  printf("= Total sim time        = %9.3f millisec      \n", clock);
  printf("=------------------------------------------------------------ \n");
  printf("= >>> Simulation results                                    - \n");
  printf("=------------------------------------------------------------ \n");
  printf("= Util of Server        = %9.3f %%            \n",
    100.0 * util(Server));
  printf("= Number of completions = %d packets          \n",
    table_cnt(Delay_table));
  printf("= Max response time     = %9.3f millisec      \n",
    table_max(Delay_table));
  printf("= Mean response time    = %9.3f millisec      \n",
    table_mean(Delay_table));
  printf("= Std Dev response time = %9.3f millisec      \n",
    sqrt(table_var(Delay_table)));
  printf("============================================================= \n");

  // Output end-of-simulation banner
  printf("*** END SIMULATION *** \n");
}

//============================================================================
//==  Function to generate IEEE workstation mix packets (Poisson arrivals)  ==
//==    - Mean packet length = 613.5 bytes                                  ==
//============================================================================
void generate(void)
{
  int      pkt_len;           // Packet length pulled from IEEE dist
  double   z;                 // Random number between 0.0 and 1.0

  create("generate");

  // Packet generation loop
  while(TRUE)
  {
    // Hold for desired interarrival time
    hold(expntl(1.0 / Lambda));

    // Select packet size in bytes from empirical IEEE distribution
    z = uniform(0.0, 1.0);
    if (z < 0.197) pkt_len = 100;
    else if (z < 0.406) pkt_len = 200;
    else if (z < 0.464) pkt_len = 300;
    else if (z < 0.498) pkt_len = 400;
    else if (z < 0.519) pkt_len = 500;
    else if (z < 0.628) pkt_len = 600;
    else if (z < 0.647) pkt_len = 700;
    else if (z < 0.649) pkt_len = 800;
    else if (z < 0.651) pkt_len = 900;
    else if (z < 0.683) pkt_len = 1000;
    else if (z < 0.879) pkt_len = 1100;
    else if (z < 0.880) pkt_len = 1200;
    else if (z < 0.881) pkt_len = 1300;
    else if (z < 0.883) pkt_len = 1400;
    else pkt_len = 1500;

    // Put the packet into the input port
    inport(pkt_len, clock);
  }
}

//============================================================================
//==  Function for input port queue                                         ==
//============================================================================
void inport(int pkt_len, double org_time)
{
  create("inport");

  // Inport queue with HOL fragment length assigned to packet length initially
  queue(Q_tail);
  HOL_frag_len = pkt_len;
  assert(queue_cnt(Q_head) == 0);
  queue(Q_head);
  HOL_frag_len = 0;
  set(Q_tail);

  // Forward to CP buffer
  outport(pkt_len, org_time);
}

//============================================================================
//==  Function for output port                                              ==
//==   - This is the CP buffer                                              ==
//============================================================================
void outport(int pkt_len, double org_time)
{
  int      i;                 // Loop counter

  create("outport");

  // Increment the length of the CP buffer
  CP_buf_len = CP_buf_len + pkt_len;
  assert(CP_buf_len <= (THRESHOLD + BLOCK_SIZE));

  // Transmit packets byte-by-byte
  reserve(Server);
  for (i=0; i<pkt_len; i++)
  {
    hold(8.0 / MEDIA_RATE);
    CP_buf_len--;
  }
  assert(CP_buf_len >= 0);
  release(Server);

  // Record the delay for this packet
  record((clock - org_time), Delay_table);
}

//============================================================================
//==  Function for poller                                                   ==
//==   - Polls input queue                                                  ==
//============================================================================
void poll(void)
{
  int      bytes_left;        // Bytes left to transfer in current block

  create("poll");

  // Polling loop
  while(TRUE)
  {
    // Set bytes_left to BLOCK_SIZE for next transfer
    bytes_left = BLOCK_SIZE;

    // Wait for a transfer to be possible
    queue(Below_thresh);

    // Transfer bytes from input queue to CP buffer
    while((HOL_frag_len > 0) && (bytes_left > 0))
    {
      if (HOL_frag_len <= bytes_left)
      {
        bytes_left = bytes_left - HOL_frag_len;
        set(Q_head);
        hold((8.0 * HOL_frag_len) / MEDIA_RATE);
      }
      else
      {
        HOL_frag_len = HOL_frag_len - bytes_left;
        hold((8.0 * bytes_left) / MEDIA_RATE);
        bytes_left = 0;
      }
    }
  }
}

//============================================================================
//==  Function for status reporter                                          ==
//==   - Report status of CP buffer to allow input queue to transfer        ==
//============================================================================
void status_reporter(void)
{
  create("status");

  // Status reporting every STATUS_LEN byte times
  while(TRUE)
  {
    // Hold STATUS_LEN byte times
    hold((8.0 * STATUS_LEN) / MEDIA_RATE);

    // Test if Below_thresh event should be set to allow transfers to occur
    if (CP_buf_len <= THRESHOLD)
      set(Below_thresh);
    else
      clear(Below_thresh);
  }
}
