//======================================================= file = inport.c =====
//=  VOQ input port for VOQ switch simulation model                           =
//=============================================================================
//=  Notes: 1)                                                                =
//=---------------------------------------------------------------------------=
//=  Contains: voq_arbit_b() - VOQ arbitration by round robin for blocks      =
//=            voq_arbit_c() - VOQ arbitration by round robin for cells       =
//=            status_reporter() - Status reporters for arbiters              =
//=            inport() - Input port containing VOQ buffers                   =
//=---------------------------------------------------------------------------=
//=  History:  KJC (01/26/02) - Genesis                                       =
//=============================================================================
#define EXTERNAL        // This is an external file (used in common.h)

//----- Include files ---------------------------------------------------------
#include <assert.h>     // Needed for assert() macro
#include "csim.h"       // Needed for CSIM
#include "common.h"     // Needed for common stuff

//=============================================================================
//==  RR poller for input port VOQ buffers                                   ==
//==   - Implements block transfers                                          ==
//=============================================================================
void voq_arbit_b(int in_id)
{
  int      out_id;              // Output port id
  int      out_poll;            // Output port id to be polled
  int      bytes_left;          // Bytes left to transfer
  int      i;                   // Loop counter

  create("voq_arbit_b");

  // Poll input port VOQ buffers forever
  out_id = 0;
  while(1)
  {
    // Wait for a transfer to be enabled
    queue(Wait_xfer[in_id]);

    // RR poll to try to find VOQs that can be transferred
    bytes_left = BLOCK_SIZE;
    for (i=0; i<N; i++)
    {
      out_poll = (out_id + i) % N;
      if ((VOQ_buf_len[in_id][out_poll] > 0) &&
          (CP_buf_len[in_id][out_poll] <= CP_THRESH))
      {
        // Transfer bytes from input queue to CP buffer
        while((HOL_frag_len[in_id][out_poll] > 0) && (bytes_left > 0))
        {
          if (HOL_frag_len[in_id][out_poll] <= bytes_left)
          {
            bytes_left = bytes_left - HOL_frag_len[in_id][out_poll];
            set(VOQ_first[in_id][out_poll]);
            hold((8.0 * HOL_frag_len[in_id][out_poll]) / MEDIA_RATE);
          }
          else
          {
            HOL_frag_len[in_id][out_poll] = HOL_frag_len[in_id][out_poll] -
                                            bytes_left;
            hold((8.0 * bytes_left) / MEDIA_RATE);
            bytes_left = 0;
          }
        }

        // Finished a transfer so increment the RR pointer
        out_id = (out_poll + 1) % N;
        if (bytes_left == 0) break;
      }
    }
  }
}

//=============================================================================
//==  RR poller for input port VOQ buffers                                   ==
//==   - Implements cell transfers                                           ==
//=============================================================================
void voq_arbit_c(int in_id)
{
  int      out_id;              // Output port id
  int      out_poll;            // Output port id to be polled
  int      i;                   // Loop counter

  create("voq_arbit_c");

  // Poll input port VOQ buffers forever
  out_id = 0;
  while(1)
  {
    // Wait for a transfer to be enabled
    queue(Wait_xfer[in_id]);

    // RR poll to try to find VOQs that can be transferred
    for (i=0; i<N; i++)
    {
      out_poll = (out_id + i) % N;
      if ((VOQ_buf_len[in_id][out_poll] > 0) &&
          (CP_buf_len[in_id][out_poll] <= CP_THRESH))
      {
        // Transfer bytes from input queue to CP buffer
        while(HOL_frag_len[in_id][out_poll] > 0)
        {
          if (HOL_frag_len[in_id][out_poll] <= CELL_SIZE)
          {
            set(VOQ_first[in_id][out_poll]);
            hold((8.0 * CELL_SIZE) / MEDIA_RATE);
          }
          else
          {
            HOL_frag_len[in_id][out_poll] = HOL_frag_len[in_id][out_poll] -
                                            CELL_SIZE;
            hold((8.0 * CELL_SIZE) / MEDIA_RATE);
          }
        }

        // Finished a transfer so increment the RR pointer
        out_id = (out_poll + 1) % N;
        break;
      }
    }
  }
}

//=============================================================================
//==  Status reporter for an input port arbiter                              ==
//=============================================================================
void status_reporter(int in_id)
{
  int      flag;                // Flag variable
  int      i;                   // Loop counter

  create("status_reporter");

  // Run forever
  while(1)
  {
    // Hold for a status update time
    hold((8.0 * STATUS_TIME) / MEDIA_RATE);

    // Test if an VOQ buffer transfer can be made
    flag = FALSE;
    for (i=0; i<N; i++)
      if ((VOQ_buf_len[in_id][i] > 0) && (CP_buf_len[in_id][i] <= CP_THRESH))
        flag = TRUE;

    // If a transfer can be made, set Wait_xfer event
    if (flag == TRUE)
      set(Wait_xfer[in_id]);
    else
      clear(Wait_xfer[in_id]);
  }
}

//=============================================================================
//==  Input port containing VOQ buffers                                      ==
//=============================================================================
void inport(int in_id, int out_id, int pkt_len, double org_time)
{
  create("inport");

  // If destination buffer is empty then set the Arrival event
  if (queue_cnt(VOQ_first[in_id][out_id]) == 0) set(State_change[in_id]);

  // Queue the arriving packet in the appropriate VOQ buffer
  VOQ_buf_len[in_id][out_id] = VOQ_buf_len[in_id][out_id] + pkt_len;
  queue(VOQ_event[in_id][out_id]);
  HOL_frag_len[in_id][out_id] = pkt_len;
  assert(queue_cnt(VOQ_first[in_id][out_id]) == 0);
  queue(VOQ_first[in_id][out_id]);
  HOL_frag_len[in_id][out_id] = 0;
  set(VOQ_event[in_id][out_id]);

  // Forward the bytes to the crosspoint buffer (as "owned" by output port)
  out[out_id](in_id, out_id, pkt_len, org_time);
  set(Xfer_cp[out_id]);
  set(Xfer_voq[in_id]);

  // Update some common variables
  VOQ_buf_len[in_id][out_id] = VOQ_buf_len[in_id][out_id] - pkt_len;

  return;
}
